/******************************************************************************
 * $Id: hilite.c,v 1.57 2001/09/05 04:01:51 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Functions for hiliting text in a text widget.
 * Notes        : Because of the huge function HiliteStringOfCode(),
 *                this code is very slow.  Slowness becomes a problem
 *                while trying to hilite a source file > 1MB because
 *                the program will appear to hang.
 ******************************************************************************/

#include "header.h"

/*
 * Color used for hilites.
 */
GdkColor colorComment  = { 0, 0x0800, 0x7000, 0x0800 };
GdkColor colorKeyword  = { 0, 0x1000, 0x1000, 0xb000 };
GdkColor colorType     = { 0, 0xb000, 0x1000, 0xc000 };
GdkColor colorSymbol   = { 0, 0xb000, 0x1000, 0x1000 };
GdkColor colorLabel    = { 0, 0x3000, 0x8000, 0x8000 };
GdkColor colorGoto     = { 0, 0xf000, 0xf000, 0x5000 };
GdkColor colorString   = { 0, 0x8000, 0x8000, 0x8000 };

/*
 * C keywords
 */
static const char keywords_c[] =
   " if"
   " else"
   " return"
   " for"
   " while"
   " do"
   " continue"
   " break"
   " switch"
   " case"
   " default"
   " sizeof"
/* " goto"  is handled/colored specially */
   " asm"
   "                 "; /* padding just in case */

/*
 * C preprocessor keywords
 */
static const char keywords_cpp[] =
   " #define"
   " #ifdef"
   " #ifndef"
   " #if"
   " #else"
   " #elif"
   " #endif"
   " #include"
   " #undef"
   " #error"
   " #pragma"
   "                 "; /* padding just in case */

/*
 * C++ keywords
 */
static const char keywords_cplusplus[] =
   " catch"
   " delete"
   " inline"
   " new"
   " operator"
   " this"
   " throw"
   " try"
   "                 "; /* padding just in case */

/*
 * ASM keywords
 */
static const char keywords_asm[] =
   /*--MASM (isn't case-sens)--*/
   " dw DW"
   " db DB"
   " dd DD"
   " align ALIGN"
   " segment SEGMENT"
   " group GROUP"
   " equ EQU"
   " assume ASSUME"
   " macro MACRO"
   " endm ENDM"
   " extern EXTERN"
   " externdef EXTERNDEF"
   " high HIGH"
   " highword HIGHWORD"
   " else ELSE"
   " endif ENDIF"
   " if IF"
   " ifdef IFDEF"
   " ife IFE"
   " ifdif IFDIF"
   " ifidn IFIDN"
   " ifb IFB"
   " ifnb IFNB"
   " include INCLUDE"
   " invoke INVOKE"
   " label LABEL"
   " low LOW"
   " lowword LOWWORD"
   " offset OFFSET"
   " org ORG"
   " proc PROC"
   " endp ENDP"
   " public PUBLIC"
   " rept REPT"
   " seg SEG"
   " size SIZE"
   " sizeof SIZEOF"
   " lengthof LENGTHOF"
   " textequ TEXTEQU"
   " byte BYTE"
   " word WORD"
   " dword DWORD"
   " ptr PTR"
   " far FAR"
   " far16 FAR16"
   " far32 FAR32"
   " near NEAR"
   " near16 NEAR16"
   " near32 NEAR32"
   /*--UNIX as/gas--*/
   /* text/data etc should be .text/.data but '.' is parsed separately */
   " text"
   " data"
   " globl"
   " global"
   " bss"
   " ascii"
   " byte"
   " char"
   " word"
   " long"
   " space"
   " section"
   " previous"
   " code16"
   "                 "; /* padding just in case */

/*
 * Perl keywords
 */
static const char keywords_perl[] =
   " if"
   " else"
   " elsif"
   " unless"
   " foreach"
   " for"
   " next"
   " last"
   " redo"
   " until"
   " while"
   " my"
   " return"
   " die"
   " exit"
   " sub"
/* " goto"  is handled/colored specially */
   " use"
   "                 "; /* padding just in case */

/*
 * C types.
 *
 * Note that including "void*" would be ineffective since alphabetic chars
 * and symbols are parsed separately.
 */
static const char types_c[] =
   " void"
   " char"
   " short"
   " int"
   " long"
   " float"
   " double"
   " const"
   " volatile"
   " register"
   " static"
   " extern"
   " signed"
   " unsigned"
   " struct"
   " union"
   " typedef"
   " enum"
   "                 "; /* padding just in case */

/*
 * C++ types.
 */
static const char types_cplusplus[] =
   " bool"
   " class"
   " friend"
   " private"
   " protected"
   " public"
   " template"
   " virtual"
   "                 "; /* padding just in case */

/*
 * Non-standard C types.
 *
 * Note that including "void*" would be ineffective since alphabetic chars
 * and symbols are parsed separately.
 */
static const char types_c_nonstd[] =
   " byte"
   " uchar"
   " uint"
   " uint8"
   " uint16"
   " uint32"
   " uint64"
   " int8"
   " int16"
   " int32"
   " int64"
   /* GLIB: */
   " gpointer"
   " gconstpointer"
   " gboolean"
   " gchar"
   " guchar"
   " gint"
   " gint8"
   " gint16"
   " gint32"
   " gint64"
   " guint"
   " guint8"
   " guint16"
   " guint32"
   " guint64"
   " glong"
   " gulong"
   " gfloat"
   "                 "; /* padding just in case */

/*
 * C/C++/other symbol chars.
 */
static const char symbols_c_other[] =
   " ="
   " =="
   " !="
   " <"
   " >"
   " <="
   " >="
   " +"
   " -"
   " *"
   " /"
   " %"
   " ||"
   " &&"
   " ?"
   " !"
   " &"
   " |"
   " ^"
   " ~"
   " ,"
   " ["
   " ]"
   " ."
   " <<"
   " >>"
   " ++"
   " --"
   " +="
   " -="
   " *="
   " /="
   " %="
   " &="
   " ^="
   " |="
   " <<="
   " >>="
   " :"
   "                 "; /* padding just in case */

/*****************************************************************************
 * Load a text widget with highlighted source code.
 *
 * This function originally highlighted C source, but it is used for other
 * languages.
 *
 * Precond : The caller must first empty and freeze the text widget.
 *
 * Parms   : pWidget
 *           Pointer to a text widget.
 *
 *           pBuffer
 *           Pointer to a buffer containing source code text.
 *           Buffer must have a trailing null char.
 *
 *           bufferSize
 *           Size of buffer (excluding trailing null char).
 *
 *           srcFileKind
 *           One of SRC_FILE_*
 *****************************************************************************/
void
HiliteLoadCode( GtkWidget* pWidget,
                char*      pBuffer,
                int        bufferSize,
                int        srcFileKind )
{
   const char*	pc;
   const char*	pcAfter;
   const char*	pcTmp1;
   const char*	pcTmp2;
   const char*	pcPrevious;
   const char*	pcPrePastWhitespace;
   const char*	pCommentAfter;
   GdkFont*		pFontComment		= NULL;
   GdkFont*		pFontKeyword		= NULL;
   GdkFont*		pFontType			= NULL;
   GdkFont*		pFontSymbol			= NULL;
   char*		pScratch;
   gboolean		useHiliteFontsOutsideComments = TRUE;
   gboolean		useHiliteFontsInsideComments  = TRUE;
   int			textWidgetSize;
   int          breakout = 0;

#define REMAINING_BUFFER_SIZE(p)	(bufferSize - (p - pBuffer))

  /*
   * Ensure the buffer is terminated by a null char,
   * otherwise the parsing of it would overrun.
   */
   g_assert( pBuffer[bufferSize] == '\0' );

  /*
   * Is buffer tiny or too big?
   * This code is so slow that trying to hilite a source file > 1MB
   * will make the program appear to hang.
   */
   if ( bufferSize < 10  ||  bufferSize > MAX_HILITE_BUFFER_LEN )
   {
      gtk_text_insert( GTK_TEXT(pWidget),
                       NULL,
                       NULL,
                       NULL,
                       pBuffer,
                       bufferSize );

      if ( bufferSize > MAX_HILITE_BUFFER_LEN )
         PrintStatusbar( "Buffer is too large to highlight." );

      return;
   }

  /*
   * Prepare fonts.
   */
   if ( useHiliteFontsInsideComments )
   {
      pFontComment = gdk_font_load( fontNameHiliteComment );
   }
   if ( useHiliteFontsOutsideComments )
   {
      pFontKeyword = gdk_font_load( fontNameHiliteKeyword );
      pFontType    = gdk_font_load( fontNameHiliteType );
      pFontSymbol  = gdk_font_load( fontNameHiliteSymbol );
   }

  /*
   * Allocate mem for a temp string as large as module buffer.
   * Will be re-used many times by lower-level functions.
   */
   pScratch = g_malloc( bufferSize );

  /*
   * Parse buffer for strings to hilite.
   */
   pcPrevious = NULL;
   for ( pc = pBuffer; pc < &pBuffer[bufferSize]; )
   {
     /*
      * Break loop if pc is stuck (due to parser bug).
      */
      if ( pc == pcPrevious )
      {
         ++breakout;
         if ( breakout > 100 )
         {
            Warning( "Hilite code is stuck." );
            WarningMalfunction();
            break;
         }
      }
      pcPrevious = pc;

     /*
      * Note: Loop should be stopped using break, not return.
      */
      g_assert( pc >= pBuffer  &&  pc < &pBuffer[bufferSize] );

     /*
      * Has end of buffer been reached?
      */
      g_assert( pc != NULL );	/* should not be NULL at this point */
      if ( pc >= &pc[bufferSize]  ||  *pc == '\0' )
         break;

     /*
      * Move past whitespace, but remember prior pointer.
      */
      pcPrePastWhitespace = pc;
      pc = (char*)PastWhitespace( pc, REMAINING_BUFFER_SIZE(pc) );

     /*
      * Does only whitespace remain?
      */
      if ( pc == NULL )
      {
         gtk_text_insert( GTK_TEXT(pWidget),
                          NULL,
                          NULL,
                          NULL,
                          pcPrePastWhitespace,
                          REMAINING_BUFFER_SIZE(pcPrePastWhitespace ) );
         break;
      }

     /*
      * Is the beginning of a comment being pointed to?
      */
      if (    (pc[0] == '/'  &&  pc[1] == '*') 	/* trailing \0 obviates checking ptr near buffer end */
           || (pc[0] == '/'  &&  pc[1] == '/')
           || (pc[0] == '#'  && (srcFileKind == SRC_FILE_PERL || srcFileKind == SRC_FILE_ASM)) /* FreeBSD asm uses # for comments */
           || (pc[0] == ';'  &&  srcFileKind == SRC_FILE_ASM)   /* MASM (MSDOS) */
           || (srcFileKind == SRC_FILE_FORTH && (pc[0]=='\\' || (pc[0]=='*' && pc[1]=='*')) ) ) /* VFORTH used '**' */
      {
         char* pCommentEndDef = cCommentEndChars;	/* for C comments */

        /*
         * Is this a one-line comment?
         */
         if (    (pc[0] == '/'  &&  pc[1] == '/')	/* C++ comment */
              || (srcFileKind == SRC_FILE_PERL)
              || (srcFileKind == SRC_FILE_ASM  && (pc[0] != '/' || pc[1] != '*') ) /* unless asm file has C comment */
              || (srcFileKind == SRC_FILE_FORTH ) )
            pCommentEndDef = "\n";

        /*
         * Parse/insert comment.
         */
         pCommentAfter = \
         HiliteParseInsertComment( pWidget,
                                   &colorComment,
                                   pFontComment,
                                   pcPrePastWhitespace,
                                   pCommentEndDef,
                                   REMAINING_BUFFER_SIZE(pcPrePastWhitespace) );

        /*
         * Was the end of the comment parsed?
         */
         if ( pCommentAfter )
         {
           /*
            * From after comment, iterate.
            */
            pc = pCommentAfter;
            continue;
         }
         else
         {
           /*
            * End of buffer must have been prematurely encountered.
            */
            break;
         }
      }

     /*
      * Is this a double-quoted string?
      */
      if ( *pc == '"' )
      {
         if ( HiliteQuotedString( pWidget,
                                  &colorString,
                                  NULL, /* default font */
                                  pcPrePastWhitespace,
                                  pc,
                                  &pBuffer[bufferSize],
                                  &pcAfter ) == FUNC_OK )
         {
            pc = pcAfter;
            continue;
         }
      }

     /*
      * Try to scoop contiguous symbol chars.
      */
      if ( ! isalpha( *pc ) )	/* perhaps non-alpha is a symbol char? */
      {
        /*
         * For each symbol char after any whitespace...
         */
         for( pcAfter = pc;
              IsCSymbolChar( *pcAfter );	/* trailing \0 obviates checking ptr near end */
              ++pcAfter )
         {
#ifdef HYPERSRC_PEDANTIC
            g_assert( pcAfter < &pBuffer[bufferSize] );
#endif
           /*
            * Break if a dquote is found unless it is the first char.
            * If dquote is first, then the dquote was earlier determined
            * as not part of a dquoted string (eg it could be '"').
            */
            if ( *pcAfter == '"'  &&  pcAfter > pc ) break;
         }

        /*
         * Were any symbol chars counted?
         */
         if ( pcAfter > pc )
         {
            gtk_text_insert( GTK_TEXT(pWidget),
                             pFontSymbol,
                             &colorSymbol,
                             NULL,
                             pcPrePastWhitespace,
                             pcAfter - pcPrePastWhitespace );
           /*
            * Move pointer, continue (stopping cases are checked above).
            */
            pc = pcAfter;
            continue;
         }

        /*
         * Fall thru.
         */
      }

     /*
      * Scoop all text until:
      * start of C comment, or symbol char (excluding dquote), or buffer end.
      */
      pcTmp1 = strstr( pc, cCommentStartChars );
      pcTmp2 = FindCSymbolChar( pc, REMAINING_BUFFER_SIZE(pc) );

      if ( pcTmp1 == NULL )					/* process lower ptr unless it is NULL */
         pcAfter = pcTmp2;
      else if ( pcTmp2 == NULL )
         pcAfter = pcTmp1;
      else
         pcAfter = min( pcTmp1, pcTmp2 );	/* neither was NULL */

      if ( pcAfter == NULL )				/* was both NULL (buffer end)? */
      {
        /*
         * Encountered buffer end.
         */
         pcAfter = &pBuffer[bufferSize];
      }

#if 0
      gtk_text_insert( GTK_TEXT(pWidget),
                       NULL,
                       NULL,
                       NULL,
                       pcPrePastWhitespace,
                       pcAfter - pcPrePastWhitespace );
#else
      HiliteStringOfCode( pWidgetTextActive,
                          pFontKeyword,
                          pFontType,
                          pFontSymbol,
                          &colorKeyword,
                          &colorType,
                          &colorSymbol,
                          pcPrePastWhitespace,
                          pcAfter - pcPrePastWhitespace,
                          pScratch,
                          srcFileKind );
#endif

     /*
      * Move pointer, continue (stopping cases are checked above).
      */
      pc = pcAfter;
   }

  /*
   * There's a bug here if the size of the text widget
   * does not equal the size of the buffer.
   */
   textWidgetSize = gtk_text_get_length( GTK_TEXT(pWidgetTextActive) );
#ifdef HYPERSRC_DEBUG
   if ( textWidgetSize != bufferSize )
   {
      DebugMsg( "HiliteLoadCCode(): gtk_text_get_length() != bufferSize in module %s", ActiveModuleName() );
      DebugMsg( "HiliteLoadCCode(): gtk_text_get_length() = %d  bufferSize = %d", textWidgetSize, bufferSize );
   }
#endif

  /*
   * Dealloc, return.
   */
   g_free( pScratch );
   return;
}

/*****************************************************************************
 * Parse a string of code (ie text outside of comment enclosures).
 *
 * Parms   : pWidget
 *           Pointer to a text widget.
 *
 *           pFont[Keyword,Type,Symbol]
 *           Hilite font (NULL can be passed).
 *
 *           pColor[Keyword,Type,Symbol]
 *           Hilite color (NULL can be passed).
 *
 *           pStr
 *			 Pointer to a string of code (uncommented).
 *
 *           strSize
 *           Size of string of code.
 *
 *           pScratch
 *           Should point to a buffer as large as corresponding module buffer.
 *
 *           srcFileKind
 *           One of SRC_FILE_*
 *
 * Returns : If a comment end was present, returns a pointer after it,
 *           or returns NULL.
 *****************************************************************************/
void
HiliteStringOfCode( GtkWidget*  pWidget,
                    GdkFont*	pFontKeyword,
                    GdkFont*	pFontType,
                    GdkFont*	pFontSymbol,
                    GdkColor*	pColorKeyword,
                    GdkColor*	pColorType,
                    GdkColor*	pColorSymbol,
                    const char* pStr,
                    int         strSize,
                    char*		pScratch,
                    int         srcFileKind )
{
   const char*	pc = NULL;
   const char*	pWord;
   int          sizeWord;

   if ( strSize == 0 ) return;
   g_assert( pStr && strSize > 0 );

   for ( pc = pStr;
         pc < &pStr[strSize];
         pc = (char*)PastWord( pWord, sizeWord ) )
   {
     /*
      * Parse next word.
      */
      sizeWord = ParseWordAlphanum( pc,
                                    &pWord,
                                    strSize - (pc - pStr) );

     /*
      * If no word was parsed (or word was too big to parse),
      * insert remainder into text widget, return.
      */
      if ( sizeWord <= 0  ||  pWord == NULL )	/* ParseWord() can return negative */
      {
         gtk_text_insert( GTK_TEXT(pWidget),
                          NULL,
                          NULL,
                          NULL,
                          pc,
                          strSize - (pc - pStr) );
         return;
      }

#ifdef HYPERSRC_PEDANTIC
g_assert( pWord >= pc  &&  pWord < &pStr[strSize] );
#endif

     /*
      * For comparing strings, extract word into a ASCIIZ string.
      */
      memcpy( pScratch, pWord, sizeWord );
      pScratch[sizeWord] = '\0';

     /*
      * Is word a keyword?
      */
      if ( /* C keywords */
              (srcFileKind == SRC_FILE_C_OR_CPLUSPLUS  &&  FindWord( keywords_c, pScratch ))

           /* C preprocessor keywords (sometimes used by non-C src files) */
           || (srcFileKind != SRC_FILE_PERL            &&  FindWord( keywords_cpp, pScratch ))

           /* C++ keywords */
           || (srcFileKind == SRC_FILE_C_OR_CPLUSPLUS  &&  FindWord( keywords_cplusplus, pScratch ))

           /* Perl keywords */
           || (srcFileKind == SRC_FILE_PERL            &&  FindWord( keywords_perl, pScratch ))

           /* ASM keywords */
           || (srcFileKind == SRC_FILE_ASM             &&  FindWord( keywords_asm,  pScratch ))
         )
      {
        /*
         * Special case: don't hilite #include <asm/..>
         */
         if ( strcmp( pScratch, "asm" ) == 0  &&  pc[3] == '/' )
            goto insert_no_hilite;

        /*
         * Insert the parsed word along with preceding whitespace.
         */
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontKeyword,
                          pColorKeyword,
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * Is word a C/C++ type?
      */
      if (     (srcFileKind == SRC_FILE_C_OR_CPLUSPLUS)
            && (    FindWord( types_c,         pScratch )
                 || FindWord( types_cplusplus, pScratch )
                 || FindWord( types_c_nonstd,  pScratch ) )
         )
      {
        /*
         * Insert the parsed word along with preceding whitespace.
         */
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontType,
                          pColorType,
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * Is word a C/C++/other symbol?
      */
      if ( FindWord( symbols_c_other, pScratch ) )
      {
        /*
         * Insert the parsed word along with preceding whitespace.
         */
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontSymbol,
                          pColorSymbol,
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * Is word "goto"?
      */
      if ( strcmp( pScratch, "goto" ) == 0 )
      {
        /*
         * Insert the parsed word along with preceding whitespace.
         */
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontKeyword,
                          &colorGoto,
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * Is word a label?
      */
      if ( pWord[sizeWord] == ':' )
      {
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontKeyword,  /* no special label font */
                          &colorLabel,   /* use special label color */
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * Word was not recognized.
      * Using normal font/color, insert the parsed word
      * along with preceding whitespace.
      */
insert_no_hilite:
      gtk_text_insert( GTK_TEXT(pWidget),
                       NULL,
                       NULL,
                       NULL,
                       pc,
                       (pWord - pc) + sizeWord );

     /*
      * Iterate...
      */
   }
}

/*****************************************************************************
 * Parse the comment that is being pointed to.
 * Insert the comment as hilited text into a text widget.
 *
 * This function is generalized.  It can accomodate C/C++ comments,
 * perhaps other langs.
 *
 * Parms   : pWidget
 *           Pointer to a text widget.
 *
 *           pFont
 *           Hilite font (NULL can be passed).
 *
 *           pColor
 *           Hilite color (NULL can be passed).
 *
 *           pCommentStartApprox
 *           This can point to either the exact start of a C comment (to "/"),
 *           or can point to whitespace that precedes the C comment.
 *
 *			 pCommentEndDef
 *           Pointer to a string that defines what the end of comment is.
 *           Eg, "/""*" or "\n" for C or C++, resp.
 *
 *           bufferSize
 *           Size of buffer, measurement based on "pCommentStart" as the start.
 *
 * Returns : If a C comment end was present, returns a pointer after it,
 *           or returns NULL.
 *****************************************************************************/
const char*
HiliteParseInsertComment( GtkWidget*  pWidget,
                          GdkColor*   pColor,
                          GdkFont*    pFont,
                          const char* pCommentStartApprox,
                          const char* pCommentEndDef,
                          int         bufferSize )
{
   char*	pCommentEnd;
   int      commentLen;

#ifdef HYPERSRC_PEDANTIC
   g_assert( pCommentStartApprox  &&  pCommentStartApprox[0]  &&  bufferSize > 0 );
#endif

  /*
   * Try to find end of comment.
   */
   pCommentEnd = strstr( pCommentStartApprox, pCommentEndDef );

  /*
   * Compute length of C comment.
   * If the C comment epilog is missing then use end-of-buffer.
   */
   if ( pCommentEnd != NULL )
      commentLen = (pCommentEnd + strlen(pCommentEndDef)) - pCommentStartApprox;
   else
      commentLen = bufferSize;

  /*
   * Insert C comment into text widget.
   */
   gtk_text_insert( GTK_TEXT(pWidget),
                    pFont,
                    pColor,
                    NULL,
                    pCommentStartApprox,
                    commentLen );

  /*
   * Return pointer to after end of comment, or NULL.
   */
   if ( pCommentEnd )
      return pCommentEnd + strlen(pCommentEndDef);
   else
      return NULL;
}

/*****************************************************************************
 * Hilite a double-quoted string.
 *
 * Parms   : pWidget
 *           Pointer to a text widget.
 *
 *           pFont
 *           Hilite font (NULL can be passed).
 *
 *           pColor
 *           Hilite color (NULL can be passed).
 *
 *           pStart
 *           Actual start of string (typically, contains leading whitespace).
 *
 *           pDquote1
 *           Pointer to first dquote.
 *
 *           pBufEnd
 *           Pointer to end of buffer.
 *
 *           pAfter (OUT)
 *           Points to next char after the ending dquote.
 *
 * Returns : FUNC_OK if successful.
 *****************************************************************************/
FUNC_X
HiliteQuotedString( GtkWidget*   pWidget,
                    GdkColor*    pColor,
                    GdkFont*     pFont,
                    const char*  pStart,
                    const char*  pDquote1,
                    const char*  pBufEnd,
                    const char** ppAfter /*OUT*/ )
{
   const char* pc;
   const char* pEol;
   const char* pDquote2;

#ifdef HYPERSRC_PEDANTIC
   if (    pStart == NULL
        || pDquote1 == NULL
        || pBufEnd == NULL
        || pDquote1 < pStart
        || pDquote1 >= pBufEnd )
   {
      WarningMalfunction();
      return FUNC_FAIL;
   }
#endif

  /*
   * Ensure first dquote is pointed to.
   */
   if ( *pDquote1 != '"' ) return FUNC_FAIL;

  /*
   * Find second dquote.  Ignore \"
   */
   pEol = ToEol( pDquote1, pBufEnd-pDquote1 );
   if ( pEol == NULL ) return FUNC_FAIL;

   pDquote2	= NULL;
   for ( pc = pDquote1+1; pc < pEol; ++pc )		/* skip first dquote */
   {
      if ( *pc == '"'  &&  *(pc-1) != '\\' )	/* ignore \" */
      {
         pDquote2 = pc;
         break;
      }
   }
   if ( pDquote2 == NULL ) return FUNC_FAIL;	/* was second dquote found? */

  /*
   * Insert dquoted string and any leading whitespace.
   */
   gtk_text_insert( GTK_TEXT(pWidget),
                    pFont,
                    pColor,
                    NULL,
                    pStart,
                    pDquote2 - pStart + 1 );

   *ppAfter = pDquote2 + 1;
   return FUNC_OK;
}

/*****************************************************************************
 * Search for an exact word in a string of words (the entire word must match).
 * Each word in the string must be separated by a leading space,
 * to avoid matches on part of a word.
 *
 * Parms   : pWords
 *           String of words (each word must be preceded by a space).
 *
 *           pSought
 *           Word that is sought (matched against words in pWords).
 *****************************************************************************/
int
FindWord( const char* pWords,
          char*       pSought )
{
   char*		pMatch;
   const char*	pWordsNext;
   int			i;
   int			soughtLen;

#ifdef HYPERSRC_PEDANTIC
   if ( pWords[0] != ' ' )
   {
      DebugMsg( "FindWords() passed an incorrect string of words." );
      return FALSE;
   }
#endif

   pWordsNext = pWords;
   soughtLen = strlen( pSought );

   for ( i = 0; i < 100; i++ ) /* prevent infinite loop */
   {
      pMatch = strstr( pWordsNext, pSought );

      if ( pMatch == NULL )
         return FALSE;

     /*
      * If the word matches exactly, there should be a leading space -and-
      * either a trailing space or null char.
      */
      if (    *(pMatch-1) == ' '
           && ( pMatch[soughtLen] == ' '  ||  pMatch[soughtLen] == 0 ) )
         return TRUE;

     /*
      * Move past the impartial match.
      */
      pWordsNext = pMatch + soughtLen;

#ifdef HYPERSRC_PEDANTIC
      if ( pWordsNext > (pWords + strlen(pWords)) )
      {
         DebugMsg( "pWordsNext > (pWords + strlen(pWords))" );
         return FALSE;
      }
#endif
   }

   DebugMsg( "Stopping infinite loop in FindString()." );
   return FALSE;
}
