/******************************************************************************
 * $Id: types.h,v 1.21 2001/02/16 06:52:18 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: structs and typedefs
 ******************************************************************************/

/******************************************************************************
 * structs/typedefs
 ******************************************************************************/

/*-----------------------------------------------------------------------------
 * Function call graph (used to construct a tree view):
 *
 * module hash table --+
 *                     |
 *                     +--> GSList of funcDef_t --+
 *                                                |
 *                                                +--> GSList of funcCall_t
 *
 *-----------------------------------------------------------------------------*/

/*
 * ############################################################
 * #   Some members point to dynamically allocated strings.   #
 * #   Some members point to static strings.                  #
 * #   Deallocating a TAG struct will be quite complicated.   #
 * ############################################################
 */
struct TAG
{
   char*	pName;			/* name of the tag */
   char*	pType;			/* type of tag, eg "func", "define", etc */
   char*	pLine;			/* which line (in string form) */
   int		lineInt;		/* which line (in integer form) */
   char*	pModule;		/* name of module the tag is in */
};
typedef struct TAG*			PTAG;

typedef struct
{
   gboolean valid;						/* true if members hold valid values */
   gfloat	adjustmentValue;			/* value from text widget's GtkAdjustment widget */
   guint    idxCursor;					/* index of the cursor within the text */
   char     module[MAX_FILENAME_LEN];
} CURSOR;

/*
 * This is stored in the module hash table.
 */
typedef struct
{
   gchar*	pModuleName;
   gpointer	pModuleBuffer;
   long		moduleSize;
 /*gboolean modified;*/		/* for later when/if editing is supported */
   GSList*	pGslistFuncDef;
   GList*   pGlistTag;		/* link in tag glist pointing to module's first tag (##INVALIDATED after tags list is sorted##) */
} HASH_ITEM_MODULE;
typedef HASH_ITEM_MODULE*	P_HASH_ITEM_MODULE;

/*
 * Defines a function call that was parsed (within a function definition).
 * A Gslist of these structs is pointed to by funcDef.pGslistFuncCall.
 *
 * Notes: If a func def/call wasn't tagged by ctags (tag lookup fails),
 *        then pTag* will point to an _unlinked_ fabricated TAG struct.
 */
typedef struct
{
   PTAG			pTagFuncCall;		/* (see above note) points to tag of called function */
} funcCall_t;

/*
 * Defines a function definition that was parsed.
 * A Gslist of these structs is pointed to by TAG.pGslistFuncDef.
 */
typedef struct
{
   PTAG			pTagFuncDef;		/* (see above note) points to function definition's tag */
   GSList*		pGslistFuncCall;
   int			lineStart;			/* line which func def begins */
   int			lineLast;			/* line which func def ends */
} funcDef_t;

typedef struct
{
   GtkWidget* pWidget;
   char       str[256];
   gboolean   focus;
   guint      fontHeight;
} focusbar_t;

/******************************************************************************
 * typedefs
 ******************************************************************************/

/*
 * For ptrs to strcmp() or strcase().
 */
typedef int FUNC_STRCMPX( const char* s1, const char *s2, size_t n);
typedef FUNC_STRCMPX* P_FUNC_STRCMPX;

/*
 * Misc. typedefs.
 */
typedef void FUNC(void);
typedef FUNC* PFUNC;
